"""
module to split a NXtomo into several
"""
import copy
import logging
from typing import Optional, Union

import h5py
import h5py._hl.selections as selection
import numpy
from silx.io.url import DataUrl
from silx.io.utils import get_data

from nxtomo.application.nxtomo import NXtomo
from nxtomo.utils.io import DatasetReader


_logger = logging.getLogger(__name__)


class NXtomoDetectorDataSplitter:
    def __init__(self, nx_tomo: NXtomo) -> None:
        """
        Splitter to split the dataset nxtomo.instrument.detector.data
        into several :class:`~nxtomo.nxobject.nxobject.NXobject`.
        This will also keep up to date rotation_angle, image_key, x_translation... datasets.
        In order to start the processing it requires a correctly formed NXtomo (same number of image_key, rotation_angle...)
        This is required for the pcotomo acquisition.

        :param :class:`~nxtomo.nxobject.nxobject.NXobject` nx_tomo: nx_tomo to be splitted
        """
        if not isinstance(nx_tomo, NXtomo):
            raise TypeError(
                f"nxtomo is expected to be an instance of {NXtomo} and not {type(nx_tomo)}"
            )
        self._nx_tomo = nx_tomo

    @property
    def nx_tomo(self) -> NXtomo:
        """
        nx_tomo to be splitted

        :param :class:`~nxtomo.nxobject.nxobject.NXobject` nx_tomo: nx_tomo to be splitted
        """
        return self._nx_tomo

    def split(
        self, data_slice: slice, nb_part: Optional[int], tomo_n: Optional[int] = None
    ) -> tuple:
        """
        split the dataset targetted to have a set of h5py.VirtualSource.

        For now we first intend to use nb_part to split the dataset. But if this one is not provided or if the
        request seems incoherent according to the number of projection and tomo_n then it will fall back on using
        tomo_n for it

        :param int nb_part: in how many contiguous dataset the instruement.detector.data must be splitted.
        :param int tomo_n: expected number of projection per NXtomo
        :raises: ValueError if the number of frame, image_key, x_translation... is incoherent.
        """
        if nb_part is not None and not isinstance(
            nb_part, (int, type(None), numpy.integer)
        ):
            raise TypeError(f"nb_part is expected to be an int not {type(nb_part)}")
        if tomo_n is not None and not isinstance(
            tomo_n, (int, type(None), numpy.integer)
        ):
            raise TypeError(f"tomo_n is expected to be an int not {type(tomo_n)}")
        if tomo_n is None and nb_part is None:
            raise ValueError("tomo_n or part_n should be provided. None provided")

        invalid_datasets = self.get_invalid_datasets()
        if len(invalid_datasets) > 0:
            _logger.warning(
                f"Some datasets have incoherent length compared to nx_tomo.instrument.detector.data length: {invalid_datasets}"
            )
        if data_slice.step not in (1, None):
            raise ValueError("slice step must be one.")
        elif nb_part is None:
            return self._split_from_tomo_n(tomo_n=tomo_n, data_slice=data_slice)
        else:
            if nb_part <= 0:
                raise ValueError(f"nb_part is expected to be >=1 not {nb_part}")
            elif nb_part == 1:
                return [
                    self.nx_tomo,
                ]
            elif (data_slice.stop - data_slice.start) % nb_part != 0 or (
                tomo_n is not None
                and ((data_slice.stop - data_slice.start) % nb_part == tomo_n)
            ):
                if tomo_n is None:
                    raise ValueError(
                        f"incoherent split requested. Request to split {(data_slice.stop - data_slice.start - 1)} slices into {nb_part} parts"
                    )
                else:
                    _logger.warning(
                        "incoherent number of slice vs (nb_part and tomo_n) to split the NXtomo. Means you probably have missing slices. Fall back on using `tomo_n` only"
                    )
                    return self._split_from_tomo_n(tomo_n=tomo_n, data_slice=data_slice)
            else:
                return self._split_from_nb_part(nb_part=nb_part, data_slice=data_slice)

    def _split_from_tomo_n(self, tomo_n: int, data_slice: slice):
        parts = []
        n_slices_remaining = data_slice.stop - data_slice.start - 1
        current_slice = data_slice
        i_part = 0
        while n_slices_remaining > 0:
            new_slice_size = tomo_n
            new_slice = slice(
                current_slice.start + new_slice_size * i_part,
                min(
                    current_slice.start + new_slice_size * (i_part + 1),
                    current_slice.stop,
                ),
                1,
            )
            nx_tomo_part = self.replace(old_slice=data_slice, new_slice=new_slice)
            parts.append(nx_tomo_part)
            n_slices_remaining -= tomo_n
            i_part += 1
        return parts

    def _split_from_nb_part(self, nb_part, data_slice: slice):
        parts = []
        current_slice = data_slice
        for i_part in range(nb_part):
            new_slice_size = (current_slice.stop - current_slice.start) // nb_part
            new_slice = slice(
                current_slice.start + new_slice_size * i_part,
                current_slice.start + new_slice_size * (i_part + 1),
                1,
            )
            nx_tomo_part = self.replace(old_slice=data_slice, new_slice=new_slice)
            parts.append(nx_tomo_part)
        return parts

    def replace(self, old_slice: slice, new_slice: slice) -> NXtomo:
        """
        replace a section of the NXtomo instrument.detector.data by a subsection of it
        """
        if not isinstance(old_slice, slice):
            raise TypeError("old_slice is expected to be a slice")
        if not isinstance(new_slice, slice):
            raise TypeError("new_slice is expected to be a slice")
        if old_slice.step not in (None, 1):
            raise ValueError("old_slice step is expected to be one")
        if new_slice.step not in (None, 1):
            raise ValueError("new_slice step is expected to be one")

        if new_slice.start < old_slice.start or new_slice.stop > old_slice.stop:
            raise ValueError(
                f"new_slice ({new_slice}) must be contained in old_slice ({old_slice})"
            )

        if old_slice.start < 0:
            raise ValueError(
                f"old_slice.start must be at least 0 not {old_slice.start}"
            )
        n_frames = self._get_n_frames()
        if n_frames is not None and old_slice.stop > n_frames:
            raise ValueError(
                f"old_slice.start must be at most {n_frames} not {old_slice.stop}"
            )

        # handles datasets other than instrument.detector.data
        result_nx_tomo = copy.deepcopy(self.nx_tomo)
        if (
            result_nx_tomo.control
            and result_nx_tomo.control.data is not None
            and result_nx_tomo.control.data.value is not None
        ):
            result_nx_tomo.control.data = numpy.concatenate(
                [
                    self.nx_tomo.control.data.value[: old_slice.start],
                    self.nx_tomo.control.data.value[new_slice],
                    self.nx_tomo.control.data.value[old_slice.stop :],
                ]
            )

        if result_nx_tomo.sample.rotation_angle is not None:
            result_nx_tomo.sample.rotation_angle = numpy.concatenate(
                [
                    self.nx_tomo.sample.rotation_angle[: old_slice.start],
                    self.nx_tomo.sample.rotation_angle[new_slice],
                    self.nx_tomo.sample.rotation_angle[old_slice.stop :],
                ]
            )

        if result_nx_tomo.sample.x_translation.value is not None:
            result_nx_tomo.sample.x_translation.value = numpy.concatenate(
                [
                    self.nx_tomo.sample.x_translation.value[: old_slice.start],
                    self.nx_tomo.sample.x_translation.value[new_slice],
                    self.nx_tomo.sample.x_translation.value[old_slice.stop :],
                ]
            )

        if result_nx_tomo.sample.y_translation.value is not None:
            result_nx_tomo.sample.y_translation.value = numpy.concatenate(
                [
                    self.nx_tomo.sample.y_translation.value[: old_slice.start],
                    self.nx_tomo.sample.y_translation.value[new_slice],
                    self.nx_tomo.sample.y_translation.value[old_slice.stop :],
                ]
            )

        if result_nx_tomo.sample.z_translation.value is not None:
            result_nx_tomo.sample.z_translation.value = numpy.concatenate(
                [
                    self.nx_tomo.sample.z_translation.value[: old_slice.start],
                    self.nx_tomo.sample.z_translation.value[new_slice],
                    self.nx_tomo.sample.z_translation.value[old_slice.stop :],
                ]
            )

        if result_nx_tomo.instrument.detector.image_key_control is not None:
            result_nx_tomo.instrument.detector.image_key_control = numpy.concatenate(
                [
                    self.nx_tomo.instrument.detector.image_key_control[
                        : old_slice.start
                    ],
                    self.nx_tomo.instrument.detector.image_key_control[new_slice],
                    self.nx_tomo.instrument.detector.image_key_control[
                        old_slice.stop :
                    ],
                ]
            )

        # handles detector.data dataset. This one is special because it can contains
        # numpy arrays (raw data), h5py.VirtualSource or DataUrl (or be None)
        det_data = self.nx_tomo.instrument.detector.data
        if det_data is None:
            pass
        elif isinstance(det_data, numpy.ndarray):
            result_nx_tomo.instrument.detector.data = numpy.concatenate(
                [
                    det_data[: old_slice.start],
                    det_data[new_slice],
                    det_data[old_slice.stop :],
                ]
            )
        elif isinstance(det_data, (tuple, list)):
            result_nx_tomo.instrument.detector.data = numpy.concatenate(
                [
                    self._get_detector_data_sub_section(slice(0, old_slice.start, 1)),
                    self._get_detector_data_sub_section(new_slice),
                    self._get_detector_data_sub_section(
                        slice(old_slice.stop, n_frames + 1, 1)
                    ),
                ]
            ).tolist()
        else:
            raise TypeError(
                f"instrument.detector.data must be a numpy array or a VirtualSource or a DataUrl. Not {type(det_data)}"
            )
        return result_nx_tomo

    def _get_detector_data_sub_section(self, section: slice) -> tuple:
        """
        return a tuple of DataUrl or h5py.VirtualSource fitting the slice requested
        """
        det_data = self.nx_tomo.instrument.detector.data
        res = []
        if section.start == section.stop:
            return ()

        def get_elmt_shape(elmt: Union[h5py.VirtualSource, DataUrl]) -> tuple:
            if isinstance(elmt, h5py.VirtualSource):
                return elmt.shape
            elif isinstance(elmt, DataUrl):
                with DatasetReader(elmt) as dataset:
                    return dataset.shape
            else:
                raise TypeError(
                    f"elmt must be a DataUrl or h5py.VirtualSource. Not {type(elmt)}"
                )

        def get_elmt_nb_frame(elmt: Union[h5py.VirtualSource, DataUrl]) -> int:
            shape = get_elmt_shape(elmt)
            if len(shape) == 3:
                return shape[0]
            elif len(shape) == 2:
                return 1
            else:
                raise ValueError(f"virtualSource: {elmt} is not 2D or 3D")

        def construct_slices_elmt_list() -> dict:
            "create a dictionary with slice as key and DataUrl or h5py.VirtualSource as value"
            slices_elmts = []

            current_index = 0
            for elmt in det_data:
                n_frame = get_elmt_nb_frame(elmt)
                slice_ = slice(current_index, current_index + n_frame, 1)
                slices_elmts.append([slice_, elmt])
                current_index += n_frame
            return slices_elmts

        def intersect(slice_1, slice_2):
            """check if the two slices intersect"""
            assert isinstance(slice_1, slice) and slice_1.step == 1
            assert isinstance(slice_2, slice) and slice_2.step == 1
            return slice_1.start < slice_2.stop and slice_1.stop > slice_2.start

        def select(
            elmt: Union[h5py.VirtualSource, DataUrl], region: slice
        ) -> Union[h5py.VirtualSource, DataUrl]:
            """select a region on the elmt.
            Can return at most the elmt itself or a region of it"""
            elmt_n_frame = get_elmt_nb_frame(elmt)
            assert elmt_n_frame != 0
            clamp_region = slice(
                max(0, region.start),
                min(elmt_n_frame, region.stop),
                1,
            )
            assert clamp_region.start != clamp_region.stop

            if isinstance(elmt, h5py.VirtualSource):
                frame_dims = elmt.shape[-2], elmt.shape[-1]
                n_frames = clamp_region.stop - clamp_region.start
                assert n_frames > 0
                shape = (n_frames, frame_dims[0], frame_dims[1])
                vs = h5py.VirtualSource(
                    path_or_dataset=elmt.path,
                    name=elmt.name,
                    shape=shape,
                )
                vs.sel = selection.select(elmt.shape, clamp_region)
                return vs
            else:
                if elmt.data_slice() is None:
                    data_slice = clamp_region
                elif isinstance(elmt.data_slice(), slice):
                    if elmt.data_slice.step not in (1, None):
                        raise ValueError("DataUrl with step !=1 are not handled")
                    else:
                        data_slice = slice(
                            elmt.data_slice.start + clamp_region.start,
                            elmt.data_slice.start + clamp_region.stop,
                            1,
                        )
                else:
                    raise TypeError(
                        f"data_slice is expected to be None or a slice. Not {type(elmt.data_slice())}"
                    )
                return DataUrl(
                    file_path=elmt.file_path(),
                    data_path=elmt.data_path(),
                    scheme=elmt.scheme(),
                    data_slice=data_slice,
                )

        for slice_raw_data, elmt in construct_slices_elmt_list():
            if intersect(section, slice_raw_data):
                res.append(
                    select(
                        elmt,
                        slice(
                            section.start - slice_raw_data.start,
                            section.stop - slice_raw_data.start,
                            1,
                        ),
                    )
                )
        return tuple(res)

    def get_invalid_datasets(self) -> dict:
        """
        return a dict of invalid dataset compare to the instrument.detector.data dataset.
        Key is the location ? path to the invalid dataset. Value is the reason of the failure.
        """
        invalid_datasets = {}
        n_frames = self._get_n_frames()

        # check rotation_angle
        if self.nx_tomo.sample.rotation_angle is not None:
            n_rotation_angles = len(self.nx_tomo.sample.rotation_angle)
            if n_rotation_angles != n_frames:
                invalid_datasets[
                    "sample/rotation_angle"
                ] = f"{n_rotation_angles} angles found when {n_frames} expected"

        # check image_key_control (force to have the same number as image_key already so only check one)
        if self.nx_tomo.instrument.detector.image_key_control is not None:
            n_image_key_control = len(
                self.nx_tomo.instrument.detector.image_key_control
            )
            if n_image_key_control != n_frames:
                invalid_datasets[
                    "instrument/detector/image_key_control"
                ] = f"{n_image_key_control} image_key_control values found when {n_frames} expected"

        # check x_translation
        if self.nx_tomo.sample.x_translation.value is not None:
            n_x_translation = len(self.nx_tomo.sample.x_translation.value)
            if n_x_translation != n_frames:
                invalid_datasets[
                    "sample/x_translation"
                ] = f"{n_x_translation} x translations found when {n_frames} expected"

        # check y_translation
        if self.nx_tomo.sample.y_translation.value is not None:
            n_y_translation = len(self.nx_tomo.sample.y_translation.value)
            if n_y_translation != n_frames:
                invalid_datasets[
                    "sample/y_translation"
                ] = f"{n_y_translation} y translations found when {n_frames} expected"

        # check z_translation
        if self.nx_tomo.sample.z_translation.value is not None:
            n_z_translation = len(self.nx_tomo.sample.z_translation.value)
            if n_z_translation != n_frames:
                invalid_datasets[
                    "sample/z_translation"
                ] = f"{n_z_translation} z translations found when {n_frames} expected"

        return invalid_datasets

    def _get_n_frames(self) -> Optional[int]:
        dataset = self.nx_tomo.instrument.detector.data
        if dataset is None:
            return None
        elif isinstance(dataset, numpy.ndarray):
            if not dataset.ndim == 3:
                raise ValueError(
                    f"nx_tomo.instrument.detector.data is expected to be 3D and not {dataset.ndim}D."
                )
            else:
                return dataset.shape[0]
        elif isinstance(dataset, (list, tuple)):
            n_frames = 0
            for dataset_elmt in dataset:
                if isinstance(dataset_elmt, h5py.VirtualSource):
                    shape = dataset_elmt.shape
                    if len(shape) == 3:
                        n_frames += dataset_elmt.shape[0]
                    elif len(shape) == 2:
                        n_frames += 1
                    else:
                        raise ValueError(
                            f"h5py.VirtualSource shape is expected to be 2D (single frame) or 3D. Not {len(shape)}D."
                        )
                elif isinstance(dataset_elmt, DataUrl):
                    data = get_data(dataset_elmt)
                    if not isinstance(data, numpy.ndarray):
                        raise TypeError(
                            f"url: {dataset_elmt.path()} is not pointing to an array"
                        )
                    elif data.ndim == 2:
                        n_frames += 1
                    elif data.ndim == 3:
                        n_frames += data.shape[0]
                    else:
                        raise ValueError(
                            f"url: {dataset_elmt.path()} is expected to be 2D or 3D. Not {dataset_elmt.ndim} D"
                        )
                else:
                    raise TypeError(
                        f"elements of {type(dataset)} must be h5py.VirtualSource) or silx.io.url.DataUrl and not {type(dataset_elmt)}"
                    )
            return n_frames
        else:
            raise TypeError(
                f"nx_tomo.instrument.detector.data type ({type(dataset)}) is not handled"
            )



"""nexus path used to define a `NXsource <https://manual.nexusformat.org/classes/base_classes/NXsource.html>`_"""


class NEXUS_SOURCE_PATH:
    NAME = "name"
    TYPE = "type"
    PROBE = "probe"


class NEXUS_SOURCE_PATH_V_1_0(NEXUS_SOURCE_PATH):
    pass


class NEXUS_SOURCE_PATH_V_1_1(NEXUS_SOURCE_PATH_V_1_0):
    pass


class NEXUS_SOURCE_PATH_V_1_2(NEXUS_SOURCE_PATH_V_1_1):
    pass


class NEXUS_SOURCE_PATH_V_1_3(NEXUS_SOURCE_PATH_V_1_2):
    pass



"""
module for handling a `nxtransformations <https://manual.nexusformat.org/classes/base_classes/nxtransformations.html#nxtransformations>`_
"""
import typing
import logging
import h5py

from copy import deepcopy

from silx.utils.proxy import docstring
from silx.io.dictdump import nxtodict

from nxtomo.nxobject.nxobject import NXobject
from nxtomo.io import HDF5File
from nxtomo.utils.transformation import (
    Transformation,
    GravityTransformation,
    get_lr_flip as _get_lr_flip,
    get_ud_flip as _get_ud_flip,
)
from nxtomo.paths.nxtomo import get_paths as get_nexus_paths

_logger = logging.getLogger(__name__)


class NXtransformations(NXobject):
    def __init__(self, node_name: str = "transformations", parent=None) -> None:
        """
        Collection of axis-based translations and rotations to describe a geometry

        For tomotools the first usage would be to allow users to provide more metadata to tag acquisition (like 'detector has been rotate' of 90 degree...)

        :param str node_name: name of the detector in the hierarchy
        :param Optional[NXObject] parent: parent in the nexus hierarchy
        """
        super().__init__(node_name, parent)
        self._set_freeze(False)
        self._transformations = dict()
        # dict with axis_name as value and Transforamtion as value. Simplify handling compared to a tuple / list / set and ensure the axis_name is unique
        self._set_freeze(True)

    @property
    def transformations(self) -> tuple:
        """
        return dict with str as key and Transformation as value
        """
        return tuple(self._transformations.values())

    @transformations.setter
    def transformations(self, transformations: tuple):
        """
        :param dict transformations: dict as [str, Transformation]
        """
        # check type
        if not isinstance(transformations, (tuple, list)):
            raise TypeError(
                f"transformations is expected to be a dict. {type(transformations)} provided instead"
            )
        for transformation in transformations:
            if not isinstance(transformation, Transformation):
                raise TypeError(
                    f"element are expected to be instances of {Transformation}. {type(transformation)} provided instead"
                )
        # convert it to a dict for conveniance
        self._transformations = {
            transformation.axis_name: transformation
            for transformation in transformations
        }

    def addTransformation(self, *args, **kwargs):
        _logger.warning("addTransformation is deprecated. Please us add_transformation")
        self.add_transformation(*args, **kwargs)

    def add_transformation(
        self, transformation: Transformation, overwrite=False, skip_if_exists=False
    ):
        """
        add a transformation to the existing one.

        :param Transformation transformation: transformation to be added
        :param bool overwrite: if a transformation with the same axis_name already exists then overwrite it
        :param bool skip_if_exists: if a transformation with the same axis_name already exists then keep the existing one
        :raises: KeyError, if a transforamtion with the same axis_name already registered
        """
        if skip_if_exists is overwrite is True:
            raise ValueError(
                "both 'skip_if_exists' and 'overwrite' set to True. Undefined behavior"
            )
        if transformation.axis_name in self._transformations:
            if overwrite:
                _logger.info(
                    "A transformation over {transformation.axis_name} is already registered. Will overwrite it"
                )
            elif skip_if_exists:
                _logger.info(
                    "A transformation over {transformation.axis_name} is already registered. Skip add"
                )
                return
            else:
                raise KeyError(
                    f"A transformation over {transformation.axis_name} is already registered. axis_name must be unique"
                )

        self._transformations[transformation.axis_name] = transformation

    def rmTransformation(self, *args, **kwargs):
        _logger.warning("rmTransformation is deprecated. Please us rm_transformation")
        self.rm_transformation(*args, **kwargs)

    def rm_transformation(self, transformation: Transformation):
        """
        remove the provided transformation to the list of existing transformation

        :param Transformation transformation: transformation to be added
        """
        self._transformations.pop(transformation.axis_name, None)

    @docstring(NXobject)
    def to_nx_dict(
        self,
        nexus_path_version: typing.Optional[float] = None,
        data_path: typing.Optional[str] = None,
        solve_empty_dependency: bool = False,
    ) -> dict:
        """
        :param bool append_gravity: If True all transformation without dependancy will be depending on a "gravity" Transformation which represent the gravity
        """
        if len(self._transformations) == 0:
            # if no transformation, avoid creating the group
            return {}
        nexus_paths = get_nexus_paths(nexus_path_version)
        transformations_nexus_paths = nexus_paths.nx_transformations_paths
        if transformations_nexus_paths is None:
            _logger.info(
                f"no TRANSFORMATIONS provided for version {nexus_path_version}"
            )
            return {}

        transformations = deepcopy(self._transformations)
        # preprocessing for gravity
        if solve_empty_dependency:
            transformations_needing_gravity = dict(
                filter(
                    lambda pair: pair[1].depends_on in (None, ""),
                    transformations.items(),
                )
            )
            if len(transformations_needing_gravity) > 0:
                gravity = GravityTransformation()
                if gravity.axis_name in transformations.keys():
                    _logger.warning(
                        f"transformations already contains a transformation named '{gravity.axis_name}'. Unable to expend transformation chain (cannot append twice gravity)"
                    )
                    gravity = None
                else:
                    transformations[gravity.axis_name] = gravity
                # update transformations needing gravity
                for transformation in transformations_needing_gravity.values():
                    transformation.depends_on = gravity.axis_name

        # dump Transformation
        nx_dict = {}
        for transformation in transformations.values():
            if not isinstance(transformation, Transformation):
                raise TypeError(
                    f"transformations are expected to be instances of {Transformation}. {type(transformation)} provided instead."
                )
            nx_dict.update(
                transformation.to_nx_dict(
                    transformations_nexus_paths=transformations_nexus_paths,
                    data_path=self.path,
                )
            )
        nx_dict[f"{self.path}@NX_class"] = "NX_TRANSFORMATIONS"
        nx_dict[f"{self.path}@units"] = "NX_TRANSFORMATION"
        return nx_dict

    @staticmethod
    def load_from_file(
        file_path: str, data_path: str, nexus_version: typing.Optional[float]
    ):
        """
        create an instance of :class:`~nxtomo.nxobject.nxtransformations,NXtransformations` and load it value from
        the given file and data path
        """
        result = NXtransformations()
        return result._load(
            file_path=file_path, data_path=data_path, nexus_version=nexus_version
        )

    def _load(
        self, file_path: str, data_path: str, nexus_version: typing.Optional[float]
    ) -> NXobject:
        """
        Create and load an NXmonitor from data on disk
        """
        nexus_paths = get_nexus_paths(nexus_version)
        transformations_nexus_paths = nexus_paths.nx_transformations_paths

        with HDF5File(file_path, mode="r") as h5f:
            if data_path == "":
                pass
            elif data_path not in h5f:
                _logger.error(
                    f"No NXtransformations found in {file_path} under {data_path} location."
                )
                return

            transformations_as_nx_dict = nxtodict(file_path, path=data_path)
            # filter attributes from the dict (as a convention dict contain '@' char)
            transformations_keys = dict(
                filter(
                    lambda a: "@" not in a[0],
                    transformations_as_nx_dict.items(),
                )
            )
            for key in transformations_keys:
                transformation = Transformation.from_nx_dict(
                    axis_name=key,
                    dict_=transformations_as_nx_dict,
                    transformations_nexus_paths=transformations_nexus_paths,
                )
                if transformation is None:
                    # if failed to load transformation (old version of nexus ?)
                    continue
                else:
                    self.add_transformation(transformation=transformation)
        return self

    @staticmethod
    @docstring(NXobject)
    def concatenate(nx_objects: tuple, node_name="transformations"):
        res = NXtransformations(node_name=node_name)
        for nx_transformations in nx_objects:
            if not isinstance(nx_transformations, NXtransformations):
                raise TypeError
            for transformation in nx_transformations.transformations:
                res.add_transformation(transformation, skip_if_exists=True)
        return res

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, NXtransformations):
            return False
        else:
            return self.transformations == __value.transformations

    @staticmethod
    def is_a_valid_group(group: h5py.Group) -> bool:
        """
        check if the group fix an NXtransformations.
        For now the only condition is to be a group and to get NXtransformations as attr
        """
        if not isinstance(group, h5py.Group):
            return False
        return group.attrs.get("NX_class", None) == "NX_TRANSFORMATIONS"

    def __len__(self):
        return len(self.transformations)


def get_lr_flip(transformations: typing.Union[tuple, NXtransformations]) -> tuple:
    """
    check along all transformations if find Transformation matching 'LRTransformation'

    return a tuple with all matching keys
    """
    if isinstance(transformations, NXtransformations):
        transformations = transformations.transformations
    return _get_lr_flip(transformations)


def get_ud_flip(transformations: typing.Union[tuple, NXtransformations]) -> tuple:
    """
    check along all transformations if find Transformation matching 'UDTransformation'

    return a tuple with all matching keys
    """
    if isinstance(transformations, NXtransformations):
        transformations = transformations.transformations
    return _get_ud_flip(transformations)



"""
module to edit, load and save `NXtomo application <https://tomotools.gitlab-pages.esrf.fr/nxtomo/>`_.
"""
from nxtomo.version import version as __version__  # noqa F401

from .application.nxtomo import NXtomo  # noqa F401



"""
module containing the definition of all the `NXobject <https://manual.nexusformat.org/classes/base_classes/NXobject.html>`_ used (and not being NXapplication)
"""
from .nxdetector import NXdetector  # noqa F401
from .nxobject import NXobject  # noqa F401
from .nxsample import NXsample  # noqa F401
from .nxsource import NXsource  # noqa F401
from .utils import concatenate  # noqa F401



"""io utils"""
import contextlib

import h5py
import logging
import traceback

from nxtomo.io import HDF5File

try:
    import hdf5plugin  # noqa F401
except ImportError:
    pass
from silx.io.url import DataUrl


class _BaseReader(contextlib.AbstractContextManager):
    def __init__(self, url: DataUrl):
        if not isinstance(url, DataUrl):
            raise TypeError(f"url should be an instance of DataUrl. Not {type(url)}")
        if url.scheme() not in ("silx", "h5py"):
            raise ValueError("Valid scheme are silx and h5py")
        if url.data_slice() is not None:
            raise ValueError(
                "Data slices are not managed. Data path should "
                "point to a bliss node (h5py.Group)"
            )
        self._url = url
        self._file_handler = None

    def __exit__(self, *exc):
        return self._file_handler.close()


class EntryReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = HDF5File(filename=self._url.file_path(), mode="r")
        if self._url.data_path() == "":
            entry = self._file_handler
        elif self._url.data_path() not in self._file_handler:
            raise KeyError(
                f"data path '{self._url.data_path()}' doesn't exists from '{self._url.file_path()}'"
            )
        else:
            entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Group):
            raise ValueError("Data path should point to a bliss node (h5py.Group)")
        return entry


class DatasetReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = HDF5File(filename=self._url.file_path(), mode="r")
        entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Dataset):
            raise ValueError(
                f"Data path ({self._url.path()}) should point to a dataset (h5py.Dataset)"
            )
        return entry


depreclog = logging.getLogger("nxtomo.DEPRECATION")

deprecache = set([])


def deprecated_warning(
    type_,
    name,
    reason=None,
    replacement=None,
    since_version=None,
    only_once=True,
    skip_backtrace_count=0,
):
    """
    Function to log a deprecation warning

    :param str type_: Nature of the object to be deprecated:
        "Module", "Function", "Class" ...
    :param name: Object name.
    :param str reason: Reason for deprecating this function
        (e.g. "feature no longer provided",
    :param str replacement: Name of replacement function (if the reason for
        deprecating was to rename the function)
    :param str since_version: First *silx* version for which the function was
        deprecated (e.g. "0.5.0").
    :param bool only_once: If true, the deprecation warning will only be
        generated one time for each different call locations. Default is true.
    :param int skip_backtrace_count: Amount of last backtrace to ignore when
        logging the backtrace
    """
    if not depreclog.isEnabledFor(logging.WARNING):
        # Avoid computation when it is not logged
        return

    msg = "%s %s is deprecated"
    if since_version is not None:
        msg += " since silx version %s" % since_version
    msg += "."
    if reason is not None:
        msg += " Reason: %s." % reason
    if replacement is not None:
        msg += " Use '%s' instead." % replacement
    msg += "\n%s"
    limit = 2 + skip_backtrace_count
    backtrace = "".join(traceback.format_stack(limit=limit)[0])
    backtrace = backtrace.rstrip()
    if only_once:
        data = (msg, type_, name, backtrace)
        if data in deprecache:
            return
        else:
            deprecache.add(data)
    depreclog.warning(msg, type_, name, backtrace)



"""
module for handling a `nxmonitor <https://manual.nexusformat.org/classes/base_classes/NXmonitor.html>`_
"""

from functools import partial
from operator import is_not
from typing import Optional, Union
import numpy

from silx.utils.proxy import docstring

from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from nxtomo.nxobject.nxobject import ElementWithUnit, NXobject
from nxtomo.utils import get_data_and_unit

from pyunitsystem import ElectricCurrentSystem


class NXmonitor(NXobject):
    def __init__(self, node_name="control", parent: Optional[NXobject] = None) -> None:
        """
        representation of `nexus NXmonitor <https://manual.nexusformat.org/classes/base_classes/NXmonitor.html>`_.
        A monitor of incident beam data.

        :param str node_name: name of the detector in the hierarchy
        :param Optional[NXObject] parent: parent in the nexus hierarchy
        """
        super().__init__(node_name=node_name, parent=parent)
        self._set_freeze(False)
        self._data = ElementWithUnit(default_unit=ElectricCurrentSystem.AMPERE)
        self._set_freeze(True)

    @property
    def data(self) -> Optional[numpy.ndarray]:
        """
        monitor data.
        In the case of NXtomo it expects to contains machine electric current for each frame
        """
        return self._data

    @data.setter
    def data(self, data: Optional[Union[numpy.ndarray, list, tuple]]):
        if isinstance(data, (tuple, list)):
            if len(data) == 0:
                data = None
            else:
                data = numpy.asarray(data)

        if isinstance(data, numpy.ndarray):
            if not data.ndim == 1:
                raise ValueError(f"data is expected to be 1D and not {data.ndim}d")
        elif not isinstance(data, type(None)):
            raise TypeError(
                f"data is expected to be None or a numpy array. Not {type(data)}"
            )
        self._data.value = data

    @docstring(NXobject)
    def to_nx_dict(
        self,
        nexus_path_version: Optional[float] = None,
        data_path: Optional[str] = None,
    ) -> dict:
        nexus_paths = get_nexus_paths(nexus_path_version)
        monitor_nexus_paths = nexus_paths.nx_monitor_paths

        nx_dict = {}
        if self.data.value is not None:
            if monitor_nexus_paths.DATA_PATH is not None:
                data_path = f"{self.path}/{monitor_nexus_paths.DATA_PATH}"
                nx_dict[data_path] = self.data.value
                nx_dict["@".join([data_path, "units"])] = str(self.data.unit)

        if nx_dict != {}:
            nx_dict[f"{self.path}@NX_class"] = "NXmonitor"
        return nx_dict

    def _load(self, file_path: str, data_path: str, nexus_version: float) -> NXobject:
        """
        Create and load an NXmonitor from data on disk
        """
        nexus_paths = get_nexus_paths(nexus_version)
        monitor_nexus_paths = nexus_paths.nx_monitor_paths
        if monitor_nexus_paths.DATA_PATH is not None:
            self.data, self.data.unit = get_data_and_unit(
                file_path=file_path,
                data_path="/".join([data_path, monitor_nexus_paths.DATA_PATH]),
                default_unit="Ampere",
            )

    @staticmethod
    @docstring(NXobject)
    def concatenate(nx_objects: tuple, node_name: str = "control"):
        # filter None obj
        nx_objects = tuple(filter(partial(is_not, None), nx_objects))
        if len(nx_objects) == 0:
            return None
        nx_monitor = NXmonitor(node_name=node_name)
        data = [
            nx_obj.data.value * nx_obj.data.unit.value
            for nx_obj in nx_objects
            if nx_obj.data.value is not None
        ]
        if len(data) > 0:
            nx_monitor.data = numpy.concatenate(data)
        return nx_monitor



"""
utils for NXobject
"""
from typing import Iterable
from nxtomo.nxobject.nxobject import NXobject


def concatenate(nx_objects: Iterable, **kwargs) -> NXobject:
    """
    concatenate a list of NXobjects

    :param Iterable nx_objects: objects to be concatenated. They are expected to be of the same type.
    :param kwargs: extra parameters
    :return: concatenated object. Of the same type of 'nx_objects'
    :rtype: :class:`~nxtomo.nxobject.nxobject.NXobject`
    """
    if len(nx_objects) == 0:
        return None
    else:
        if not isinstance(nx_objects[0], NXobject):
            raise TypeError("nx_objects are expected to be instances of NXobject")
        return type(nx_objects[0]).concatenate(nx_objects=nx_objects, **kwargs)



import numpy
import pytest

from nxtomo.nxobject.nxtransformations import NXtransformations
from nxtomo.utils.transformation import (
    Transformation,
    TransformationAxis,
)


def test_nx_transforamtions(tmp_path):
    """test creation and saving of an NXtransformations"""
    nx_transformations_1 = NXtransformations()
    with pytest.raises(TypeError):
        nx_transformations_1.transformations = 12
    with pytest.raises(TypeError):
        nx_transformations_1.transformations = {12: 12}

    translation_along_x = Transformation(
        axis_name="tx",
        value=9.6,
        transformation_type="translation",
        vector=TransformationAxis.AXIS_X,
    )
    nx_transformations_1.add_transformation(
        transformation=translation_along_x,
    )
    rotation_along_z = Transformation(
        axis_name="rz",
        value=90,
        transformation_type="rotation",
        vector=TransformationAxis.AXIS_Z,
    )
    rotation_along_z.offset = (12.0, 0, 0)
    assert numpy.array_equal(rotation_along_z.offset, numpy.array([12.0, 0, 0]))
    rotation_along_z.units = "degree"

    rotation_along_z.depends_on = "tx"
    assert rotation_along_z.depends_on == "tx"
    with pytest.raises(AttributeError):
        rotation_along_z.vector = TransformationAxis.AXIS_Z
    assert rotation_along_z.vector == (0, 0, 1)

    nx_transformations_1.add_transformation(
        rotation_along_z,
    )
    assert len(nx_transformations_1.transformations) == 2

    assert nx_transformations_1.to_nx_dict(data_path="") == {
        # ty specifics
        "tx": 9.6,
        "tx@transformation_type": "translation",
        "tx@units": "m",
        "tx@vector": (1, 0, 0),
        "tx@offset": (0, 0, 0),
        # tx specifics
        "rz": 90,
        "rz@depends_on": "tx",
        "rz@offset": (12.0, 0, 0),
        "rz@transformation_type": "rotation",
        "rz@units": "degree",
        "rz@vector": (0, 0, 1),
        # class attributes
        "@NX_class": "NX_TRANSFORMATIONS",
        "@units": "NX_TRANSFORMATION",
    }

    # check solving empty dependancy
    assert nx_transformations_1.to_nx_dict(
        data_path="", solve_empty_dependency=True
    ) == {
        # ty specifics
        "tx": 9.6,
        "tx@transformation_type": "translation",
        "tx@units": "m",
        "tx@vector": (1, 0, 0),
        "tx@offset": (0, 0, 0),
        "tx@depends_on": "gravity",
        # tx specifics
        "rz": 90,
        "rz@depends_on": "tx",
        "rz@offset": (12.0, 0, 0),
        "rz@transformation_type": "rotation",
        "rz@units": "degree",
        "rz@vector": (0, 0, 1),
        # gravity
        "gravity": numpy.nan,
        "gravity@offset": (0, 0, 0),
        "gravity@transformation_type": "translation",
        "gravity@units": "m/s2",
        "gravity@vector": (0, 0, -1),
        # class attributes
        "@NX_class": "NX_TRANSFORMATIONS",
        "@units": "NX_TRANSFORMATION",
    }

    nx_transformations_2 = NXtransformations()
    nx_transformations_2.transformations = (
        Transformation("rx", 60, "rotation", vector=TransformationAxis.AXIS_X),
        Transformation("rz", -60, "rotation", vector=TransformationAxis.AXIS_Z),
    )

    assert NXtransformations.concatenate(
        [nx_transformations_2, nx_transformations_1]
    ).transformations == (
        Transformation("rx", 60, "rotation", vector=TransformationAxis.AXIS_X),
        Transformation("rz", -60, "rotation", vector=TransformationAxis.AXIS_Z),
        translation_along_x,
    )

    assert NXtransformations.concatenate(
        [nx_transformations_1, nx_transformations_2]
    ).transformations != (
        translation_along_x,
        Transformation("rx", 60, "rotation", vector=TransformationAxis.AXIS_X),
        Transformation("rz", -60, "rotation", vector=TransformationAxis.AXIS_Z),
    )

    # save NXtransformation to file and load it
    output_file_path = str(tmp_path / "test_nxtransformations.nx")
    nx_transformations_2.save(output_file_path, "transformations")
    assert len(nx_transformations_2.transformations) == 2

    # test backward compatibility
    loaded_transformations = NXtransformations()._load(
        output_file_path, "transformations", 1.2
    )
    assert isinstance(loaded_transformations, NXtransformations)
    assert len(loaded_transformations.transformations) == 0

    # test backward compatibility
    loaded_transformations = NXtransformations()._load(
        output_file_path, "transformations", 1.3
    )
    assert isinstance(loaded_transformations, NXtransformations)
    assert len(loaded_transformations.transformations) == 2
    assert loaded_transformations == nx_transformations_2

    output_file_path_2 = str(tmp_path / "test_nxtransformations.nx")
    nx_transformations_2.save(output_file_path_2, "/entry/toto/transformations")

    loaded_transformations = NXtransformations()._load(
        output_file_path_2, "/entry/toto/transformations", 1.3
    )
    assert isinstance(loaded_transformations, NXtransformations)
    assert len(loaded_transformations.transformations) == 2
    assert loaded_transformations == nx_transformations_2



import pytest

from nxtomo.nxobject.nxsource import NXsource


def test_nx_source():
    """test creation and saving of an nxsource"""
    nx_source = NXsource()
    with pytest.raises(TypeError):
        nx_source.name = 12
    nx_source.name = "my source"

    with pytest.raises(AttributeError):
        nx_source.source_name = "test"

    with pytest.raises(ValueError):
        nx_source.type = "toto"
    nx_source.type = "Synchrotron X-ray Source"
    str(nx_source)
    nx_source.type = None
    str(nx_source)

    assert nx_source.probe is None
    nx_source.probe = "neutron"
    assert nx_source.probe.value == "neutron"
    with pytest.raises(ValueError):
        nx_source.probe = 12

    assert isinstance(nx_source.to_nx_dict(), dict)

    # check we can't set undefined attributes
    with pytest.raises(AttributeError):
        nx_source.test = 12

    # test some concatenation
    nx_source_concatenate = NXsource.concatenate([nx_source, nx_source])
    assert nx_source_concatenate.name == "my source"
    assert nx_source_concatenate.type is None
    assert nx_source_concatenate.probe.value == "neutron"



import os
from tempfile import TemporaryDirectory

import h5py
import numpy
import pytest
from silx.io.url import DataUrl

from nxtomo.application.nxtomo import NXtomo
from nxtomo.utils.detectorsplitter import NXtomoDetectorDataSplitter


def test_get_invalid_datasets():
    """test the NXtomoDetectorDataSplitter get_invalid_datasets function"""
    nx_tomo = NXtomo("test")
    n_frames = 10
    nx_tomo.instrument.detector.data = numpy.random.random(100 * 100 * 10).reshape(
        [n_frames, 100, 100]
    )
    splitter = NXtomoDetectorDataSplitter(nx_tomo=nx_tomo)
    assert len(splitter.get_invalid_datasets()) == 0

    # test rotation angle
    nx_tomo.sample.rotation_angle = [12, 13]
    assert len(splitter.get_invalid_datasets()) == 1
    nx_tomo.sample.rotation_angle = [0] * n_frames
    assert len(splitter.get_invalid_datasets()) == 0

    # test image_key_control
    nx_tomo.instrument.detector.image_key_control = [0]
    assert len(splitter.get_invalid_datasets()) == 1
    nx_tomo.instrument.detector.image_key_control = [0] * n_frames
    assert len(splitter.get_invalid_datasets()) == 0

    # test x_translation
    nx_tomo.sample.x_translation = [0]
    assert len(splitter.get_invalid_datasets()) == 1
    nx_tomo.sample.x_translation = [0] * n_frames
    assert len(splitter.get_invalid_datasets()) == 0

    # test y_translation
    nx_tomo.sample.y_translation = [0]
    assert len(splitter.get_invalid_datasets()) == 1
    nx_tomo.sample.y_translation = [0] * n_frames
    assert len(splitter.get_invalid_datasets()) == 0

    # test z_translation
    nx_tomo.sample.z_translation = [0]
    assert len(splitter.get_invalid_datasets()) == 1
    nx_tomo.sample.z_translation = [0] * n_frames
    assert len(splitter.get_invalid_datasets()) == 0


def test_spliter_raw_data():
    """test the splitter on a simple non virtual h5py Dataset"""
    nx_tomo = NXtomo("test_raw_data")
    n_frames = 20
    nx_tomo.instrument.detector.data = numpy.random.random(
        100 * 100 * n_frames
    ).reshape([n_frames, 100, 100])
    nx_tomo.sample.rotation_angle = [0, 12]
    splitter = NXtomoDetectorDataSplitter(nx_tomo=nx_tomo)
    # check incoherent number of rotation
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 100, 1), nb_part=2)
    nx_tomo.sample.rotation_angle = numpy.linspace(0, 180, num=n_frames, endpoint=False)
    # check slice nb_part < 0
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 100, 1), nb_part=-1)
    # check slice step != 1
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 100, 2), nb_part=2)
    # check incoherent number of frames
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 99, 2), nb_part=2)

    # check x translation
    nx_tomo.sample.x_translation = [0, 12]
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 100, 1), nb_part=2)
    nx_tomo.sample.x_translation = numpy.random.random(n_frames)
    nx_tomo.sample.y_translation = numpy.random.random(n_frames)
    nx_tomo.sample.z_translation = numpy.random.random(n_frames)
    # check image key
    nx_tomo.instrument.detector.image_key_control = [0, 2]
    with pytest.raises(ValueError):
        splitter.split(data_slice=slice(0, 100, 1), nb_part=2)
    nx_tomo.instrument.detector.image_key_control = [
        numpy.random.randint(low=-1, high=2) for i in range(n_frames)
    ]

    assert splitter.split(data_slice=slice(0, 100, 1), nb_part=1) == [
        nx_tomo,
    ]

    # check error if request to split a region bigger that the one (100 vs n_frames)
    with pytest.raises(ValueError):
        splitted_nx_tomo = splitter.split(data_slice=slice(0, 100, 1), nb_part=2)

    splitted_nx_tomo = splitter.split(data_slice=slice(0, 20, 1), nb_part=2)
    assert len(splitted_nx_tomo) == 2
    s_nx_tomo_1, s_nx_tomo_2 = splitted_nx_tomo
    # chek rotation_angle
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.sample.rotation_angle,
        nx_tomo.sample.rotation_angle[0 : n_frames // 2],
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.sample.rotation_angle,
        nx_tomo.sample.rotation_angle[n_frames // 2 :],
    )
    # check image key and image key
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.instrument.detector.image_key_control,
        nx_tomo.instrument.detector.image_key_control[0 : n_frames // 2],
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.instrument.detector.image_key_control,
        nx_tomo.instrument.detector.image_key_control[n_frames // 2 :],
    )

    # chek x translation
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.sample.x_translation.value,
        nx_tomo.sample.x_translation.value[0 : n_frames // 2],
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.sample.x_translation.value,
        nx_tomo.sample.x_translation.value[n_frames // 2 :],
    )

    # chek y translation
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.sample.y_translation.value,
        nx_tomo.sample.y_translation.value[0 : n_frames // 2],
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.sample.y_translation.value,
        nx_tomo.sample.y_translation.value[n_frames // 2 :],
    )
    # chek z translation
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.sample.z_translation.value,
        nx_tomo.sample.z_translation.value[0 : n_frames // 2],
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.sample.z_translation.value,
        nx_tomo.sample.z_translation.value[n_frames // 2 :],
    )
    # check detector data
    numpy.testing.assert_array_equal(
        s_nx_tomo_1.instrument.detector.data,
        nx_tomo.instrument.detector.data[0 : n_frames // 2],
    )
    numpy.testing.assert_array_equal(
        s_nx_tomo_2.instrument.detector.data,
        nx_tomo.instrument.detector.data[n_frames // 2 :],
    )


def test_spliter_virtual_sources_1():
    """
    test the splitter on a simulated h5py virtual dataset composed of two Virtual source.
    It must have the same virtual source in the two splitted NXtomo
    test the splitter on a h5py virtual dataset
    rotation_angle, [W]_translation and image_key datasets are handle are always numpy arrays
    not pointing to any external ressources. This is only the case for
    detector.data so this is the only dataset to test here
    """
    nx_tomo = NXtomo("test_raw_data")
    nx_tomo.instrument.detector.data = [
        h5py.VirtualSource("path_to_dataset_1", name="dataset_1", shape=[10, 100, 100]),
        h5py.VirtualSource("path_to_dataset_2", name="dataset_2", shape=[10, 100, 100]),
    ]
    splitter = NXtomoDetectorDataSplitter(nx_tomo=nx_tomo)
    splitted_nx_tomo = splitter.split(data_slice=slice(0, 20, 1), nb_part=2)
    assert len(splitted_nx_tomo) == 2
    s_nx_tomo_1, s_nx_tomo_2 = splitted_nx_tomo
    det_dataset_1 = s_nx_tomo_1.instrument.detector.data
    det_dataset_2 = s_nx_tomo_2.instrument.detector.data
    assert len(det_dataset_1) == 1
    assert len(det_dataset_2) == 1

    det_dataset_vs1 = det_dataset_1[0]
    det_dataset_vs2 = det_dataset_2[0]
    assert isinstance(det_dataset_vs1, h5py.VirtualSource)
    assert det_dataset_vs1.path == "path_to_dataset_1"
    assert det_dataset_vs1.shape == (10, 100, 100)

    assert isinstance(det_dataset_vs2, h5py.VirtualSource)
    assert det_dataset_vs2.path == "path_to_dataset_2"
    assert det_dataset_vs2.shape == (10, 100, 100)


def test_spliter_virtual_sources_2():
    """
    test the splitter on a h5py virtual dataset composed of a single Virtual source.
    Must split this one into two VirtualSource
    rotation_angle, [W]_translation and image_key datasets are handle are always numpy arrays
    not pointing to any external ressources. This is only the case for
    detector.data so this is the only dataset to test here
    """
    nx_tomo = NXtomo("test_raw_data")
    nx_tomo.instrument.detector.data = [
        h5py.VirtualSource(
            "path_to_dataset", name="path_to_dataset", shape=[20, 100, 100]
        ),
    ]
    splitter = NXtomoDetectorDataSplitter(nx_tomo=nx_tomo)
    splitted_nx_tomo = splitter.split(data_slice=slice(0, 20, 1), nb_part=2)
    assert len(splitted_nx_tomo) == 2

    splitted_nx_tomo = splitter.split(data_slice=slice(0, 20, 1), nb_part=4)
    assert len(splitted_nx_tomo) == 4


def test_spliter_virtual_sources_3():
    """
    test the splitter on a concrete h5py virtual dataset
    rotation_angle, [W]_translation and image_key datasets are handle are always numpy arrays
    not pointing to any external ressources. This is only the case for
    detector.data so this is the only dataset to test here
    """
    n_file = 5
    n_frame_per_file = 20
    layout = h5py.VirtualLayout(
        shape=(n_file * n_frame_per_file, 100, 100), dtype=float
    )
    with TemporaryDirectory() as folder:
        for i_file in range(n_file):
            file_path = os.path.join(folder, f"file{i_file}.hdf5")
            data_path = f"path_to_dataset_{i_file}"
            with h5py.File(file_path, mode="w") as h5f:
                if i_file == 0:
                    data = numpy.ones([n_frame_per_file, 100, 100])
                elif i_file == n_file - 1:
                    data = numpy.ones([n_frame_per_file, 100, 100]) * 2
                else:
                    start = i_file * 1000.0
                    stop = i_file * 1000.0 + (n_frame_per_file * 100 * 100)
                    data = numpy.arange(start, stop).reshape(n_frame_per_file, 100, 100)
                h5f[data_path] = data
                vs = h5py.VirtualSource(h5f[data_path])
                layout[i_file * n_frame_per_file : (i_file + 1) * n_frame_per_file] = vs

        master_file = os.path.join(folder, "master_file.hdf5")
        with h5py.File(master_file, mode="w") as h5f:
            h5f.create_virtual_dataset("data", layout)
            original_data = h5f["data"][()]

        nx_tomo = NXtomo("entry0000")
        with h5py.File(master_file, mode="r") as h5f:
            vs_ = []
            for vs_info in h5f["data"].virtual_sources():
                vs_.append(
                    h5py.VirtualSource(
                        vs_info.file_name,
                        vs_info.dset_name,
                        shape=(n_frame_per_file, 100, 100),
                    )
                )

            nx_tomo.instrument.detector.data = vs_

        splitter = NXtomoDetectorDataSplitter(nx_tomo=nx_tomo)
        data_slice = slice(10, n_frame_per_file * n_file - 10, 1)
        splitted_nx_tomo = splitter.split(data_slice=data_slice, nb_part=2)
        assert len(splitted_nx_tomo) == 2
        # check the two dataset created
        s_nx_tomo_1, s_nx_tomo_2 = splitted_nx_tomo
        output_file_1 = os.path.join(folder, "output_file_1.nx")

        # data must contains a common section between the two nxtomo: the first 10 and last 10 frames
        # then the rest must be splitted between the two NXtomo
        assert len(s_nx_tomo_1.instrument.detector.data) == 5
        assert s_nx_tomo_1.instrument.detector.data[0].shape[0] == 10
        assert s_nx_tomo_1.instrument.detector.data[1].shape[0] == 10
        assert s_nx_tomo_1.instrument.detector.data[2].shape[0] == 20
        assert s_nx_tomo_1.instrument.detector.data[3].shape[0] == 10
        assert s_nx_tomo_1.instrument.detector.data[4].shape[0] == 10
        s_nx_tomo_1.save(output_file_1, "entry0000")

        output_file_2 = os.path.join(folder, "output_file_2.nx")
        assert len(s_nx_tomo_2.instrument.detector.data) == 5
        assert s_nx_tomo_2.instrument.detector.data[0].shape[0] == 10
        assert s_nx_tomo_2.instrument.detector.data[1].shape[0] == 10
        assert s_nx_tomo_2.instrument.detector.data[2].shape[0] == 20
        assert s_nx_tomo_2.instrument.detector.data[3].shape[0] == 10
        assert s_nx_tomo_2.instrument.detector.data[4].shape[0] == 10
        s_nx_tomo_2.save(output_file_2, "entry0000")

        # check final datasets are correctly formed
        with h5py.File(output_file_1, mode="r") as h5f:
            nx_1_data = h5f["entry0000/instrument/detector/data"][()]
        assert nx_1_data.shape[0] == 60

        # check final datasets are correctly formed
        with h5py.File(output_file_2, mode="r") as h5f:
            nx_2_data = h5f["entry0000/instrument/detector/data"][()]
        assert nx_2_data.shape[0] == 60

        # first 10 frames (common between the three nxtomo)
        numpy.testing.assert_array_equal(
            nx_1_data[0:10],
            nx_2_data[0:10],
        )
        numpy.testing.assert_array_equal(
            nx_1_data[0:10],
            original_data[0:10],
        )

        # last 10 frames (common between the three nxtomo)
        numpy.testing.assert_array_equal(
            nx_1_data[-10:],
            nx_2_data[-10:],
        )
        numpy.testing.assert_array_equal(
            nx_1_data[-10:],
            original_data[-10:],
        )

        # test nx_1_data unique region
        numpy.testing.assert_array_equal(
            nx_1_data[10:50],
            original_data[10:50],
        )

        # test nx_2_data unique region
        numpy.testing.assert_array_equal(
            nx_2_data[10:50],
            original_data[50:90],
        )


def test_spliter_data_url():
    """
    test the splitter on a list of DataUrl
    rotation_angle, [W]_translation and image_key datasets are handle are always numpy arrays
    not pointing to any external ressources. This is only the case for
    detector.data so this is the only dataset to test here
    """
    urls = []
    n_frame_per_file = 20
    n_file = 5
    original_data = []
    with TemporaryDirectory() as folder:
        for i_file in range(n_file):
            file_path = os.path.join(folder, f"file{i_file}.hdf5")
            data_path = f"path_to_dataset_{i_file}"
            with h5py.File(file_path, mode="w") as h5f:
                if i_file == 0:
                    data = numpy.ones([n_frame_per_file, 100, 100])
                elif i_file == n_file - 1:
                    data = numpy.ones([n_frame_per_file, 100, 100]) * 2
                else:
                    start = i_file * 1000.0
                    stop = i_file * 1000.0 + (n_frame_per_file * 100 * 100)
                    data = numpy.arange(start, stop).reshape(n_frame_per_file, 100, 100)
                h5f[data_path] = data
                original_data.append(data)

            urls.append(
                DataUrl(
                    file_path=file_path,
                    data_path=data_path,
                    scheme="silx",
                )
            )

        original_data = numpy.concatenate(original_data)

        nx_tomo = NXtomo("entry0000")
        nx_tomo.instrument.detector.data = urls

        splitter = NXtomoDetectorDataSplitter(nx_tomo=nx_tomo)
        data_slice = slice(10, n_frame_per_file * n_file - 10, 1)
        data_slice = slice(10, n_frame_per_file * n_file - 10, 1)
        splitted_nx_tomo = splitter.split(data_slice=data_slice, nb_part=2)
        assert len(splitted_nx_tomo) == 2
        # check the two dataset created
        s_nx_tomo_1, s_nx_tomo_2 = splitted_nx_tomo
        output_file_1 = os.path.join(folder, "output_file_1.nx")

        # data must contains a common section between the two nxtomo: the first 10 and last 10 frames
        # then the rest must be splitted between the two NXtomo
        def n_elmt(slice_):
            return slice_.stop - slice_.start

        assert len(s_nx_tomo_1.instrument.detector.data) == 5
        assert n_elmt(s_nx_tomo_1.instrument.detector.data[0].data_slice()) == 10
        assert n_elmt(s_nx_tomo_1.instrument.detector.data[1].data_slice()) == 10
        assert n_elmt(s_nx_tomo_1.instrument.detector.data[2].data_slice()) == 20
        assert n_elmt(s_nx_tomo_1.instrument.detector.data[3].data_slice()) == 10
        assert n_elmt(s_nx_tomo_1.instrument.detector.data[4].data_slice()) == 10
        s_nx_tomo_1.save(output_file_1, "entry0000")

        output_file_2 = os.path.join(folder, "output_file_2.nx")
        assert len(s_nx_tomo_2.instrument.detector.data) == 5
        assert n_elmt(s_nx_tomo_2.instrument.detector.data[0].data_slice()) == 10
        assert n_elmt(s_nx_tomo_2.instrument.detector.data[1].data_slice()) == 10
        assert n_elmt(s_nx_tomo_2.instrument.detector.data[2].data_slice()) == 20
        assert n_elmt(s_nx_tomo_2.instrument.detector.data[3].data_slice()) == 10
        assert n_elmt(s_nx_tomo_2.instrument.detector.data[4].data_slice()) == 10
        s_nx_tomo_2.save(output_file_2, "entry0000")

        # check final datasets are correctly formed
        with h5py.File(output_file_1, mode="r") as h5f:
            nx_1_data = h5f["entry0000/instrument/detector/data"][()]
        assert nx_1_data.shape[0] == 60

        # check final datasets are correctly formed
        with h5py.File(output_file_2, mode="r") as h5f:
            nx_2_data = h5f["entry0000/instrument/detector/data"][()]
        assert nx_2_data.shape[0] == 60

        # first 10 frames (common between the three nxtomo)
        numpy.testing.assert_array_equal(
            nx_1_data[0:10],
            nx_2_data[0:10],
        )
        numpy.testing.assert_array_equal(
            nx_1_data[0:10],
            original_data[0:10],
        )

        # last 10 frames (common between the three nxtomo)
        numpy.testing.assert_array_equal(
            nx_1_data[-10:],
            nx_2_data[-10:],
        )
        numpy.testing.assert_array_equal(
            nx_1_data[-10:],
            original_data[-10:],
        )

        # test nx_1_data unique region
        numpy.testing.assert_array_equal(
            nx_1_data[10:50],
            original_data[10:50],
        )

        # test nx_2_data unique region
        numpy.testing.assert_array_equal(
            nx_2_data[10:50],
            original_data[50:90],
        )


def test_spliter_missing_projections():
    """
    If some projections are missing and if nb_turn cannot be used then 'tomo_n' will be used instead
    """
    urls = []
    n_frame_per_file = 20
    n_file = 5
    original_data = []
    with TemporaryDirectory() as folder:
        for i_file in range(n_file):
            file_path = os.path.join(folder, f"file{i_file}.hdf5")
            data_path = f"path_to_dataset_{i_file}"
            with h5py.File(file_path, mode="w") as h5f:
                if i_file == 0:
                    data = numpy.ones([n_frame_per_file, 100, 100])
                elif i_file == n_file - 1:
                    data = numpy.ones([n_frame_per_file, 100, 100]) * 2
                else:
                    start = i_file * 1000.0
                    stop = i_file * 1000.0 + (n_frame_per_file * 100 * 100)
                    data = numpy.arange(start, stop).reshape(n_frame_per_file, 100, 100)
                h5f[data_path] = data
                original_data.append(data)

            urls.append(
                DataUrl(
                    file_path=file_path,
                    data_path=data_path,
                    scheme="silx",
                )
            )

        original_data = numpy.concatenate(original_data)

        nx_tomo = NXtomo("entry0000")
        nx_tomo.instrument.detector.data = urls

        splitter = NXtomoDetectorDataSplitter(nx_tomo=nx_tomo)
        data_slice = slice(0, 100, 1)
        data_slice = slice(0, 100, 1)
        splitted_nx_tomo = splitter.split(data_slice=data_slice, nb_part=2)
        assert len(splitted_nx_tomo) == 2

        splitted_nx_tomo = splitter.split(
            data_slice=data_slice, nb_part=None, tomo_n=20
        )
        assert len(splitted_nx_tomo) == 5

        splitted_nx_tomo = splitter.split(
            data_slice=data_slice, nb_part=None, tomo_n=40
        )
        assert len(splitted_nx_tomo) == 3

        splitted_nx_tomo = splitter.split(
            data_slice=data_slice, nb_part=None, tomo_n=65
        )
        assert len(splitted_nx_tomo) == 2



# -- Project information -----------------------------------------------------

project = "nxtomo"
copyright = "2023, ESRF"
author = "P.Paleo, H.Payno, A. Mirone, J.Lesaint"

# The full version, including alpha/beta/rc tags
release = "1.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.autosummary",
    "nbsphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "nature"

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

html_logo = "img/nxtomo.png"

# autosummary options
autosummary_generate = True

autodoc_default_flags = [
    "members",
    "undoc-members",
    "show-inheritance",
]



"""
some io utils to handle `nexus <https://manual.nexusformat.org/index.html>`_ and `hdf5 <https://www.hdfgroup.org/solutions/hdf5/>`_ with `h5py <https://www.h5py.org/>`_
"""
import errno
import logging
import os
import traceback
from contextlib import ExitStack, contextmanager
import h5py._hl.selections as selection
from silx.io.url import DataUrl

import h5py
import threading


HASSWMR = h5py.version.hdf5_version_tuple >= h5py.get_config().swmr_min_hdf5_version
_logger = logging.getLogger(__name__)


class SharedLockPool:
    """
    Allows to acquire locks identified by name (hashable type) recursively.
    """

    def __init__(self):
        self.__locks = {}
        self.__locks_mutex = threading.Semaphore(value=1)

    def __len__(self):
        return len(self.__locks)

    @property
    def names(self):
        return list(self.__locks.keys())

    @contextmanager
    def _modify_locks(self):
        self.__locks_mutex.acquire()
        try:
            yield self.__locks
        finally:
            self.__locks_mutex.release()

    @contextmanager
    def acquire(self, name):
        with self._modify_locks() as locks:
            lock = locks.get(name, None)
            if lock is None:
                locks[name] = lock = threading.RLock()
        lock.acquire()
        try:
            yield
        finally:
            lock.release()
            with self._modify_locks() as locks:
                if name in locks:
                    locks.pop(name)

    @contextmanager
    def acquire_context_creation(self, name, contextmngr, *args, **kwargs):
        """
        Acquire lock only during context creation.

        This can be used for example to protect the opening of a file
        but not hold the lock while the file is open.
        """
        with ExitStack() as stack:
            with self.acquire(name):
                ret = stack.enter_context(contextmngr(*args, **kwargs))
            yield ret


class HDF5File(h5py.File):
    """
    File to secure reading and writing within h5py

    code originally from bliss.nexus_writer_service.io.nexus
    """

    _LOCKPOOL = SharedLockPool()

    def __init__(self, filename, mode, enable_file_locking=None, swmr=None, **kwargs):
        """
        :param str filename:
        :param str mode:
        :param bool enable_file_locking: by default it is disabled for `mode=='r'`
                                         and enabled in all other modes
        :param bool swmr: when not specified: try both modes when `mode=='r'`
        :param **kwargs: see `h5py.File.__init__`
        """
        if mode not in ("r", "w", "w-", "x", "a"):
            raise ValueError(f"invalid mode {mode}")

        with self._protect_init(filename):
            # https://support.hdfgroup.org/HDF5/docNewFeatures/SWMR/Design-HDF5-FileLocking.pdf
            if not HASSWMR and swmr:
                swmr = False
            libver = kwargs.get("libver")
            if swmr:
                kwargs["libver"] = "latest"
            if enable_file_locking is None:
                enable_file_locking = mode != "r"
            old_file_locking = os.environ.get("HDF5_USE_FILE_LOCKING", None)
            if enable_file_locking:
                os.environ["HDF5_USE_FILE_LOCKING"] = "TRUE"
            else:
                os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
            kwargs["track_order"] = True
            try:
                super().__init__(filename, mode=mode, swmr=swmr, **kwargs)
                if mode != "r" and swmr:
                    # Try setting writing in SWMR mode
                    try:
                        self.swmr_mode = True
                    except Exception:
                        pass
            except OSError as e:
                if (
                    swmr is not None
                    or mode != "r"
                    or not HASSWMR
                    or not isErrno(e, errno.EAGAIN)
                ):
                    raise
                # Try reading with opposite SWMR mode
                swmr = not swmr
                if swmr:
                    kwargs["libver"] = "latest"
                else:
                    kwargs["libver"] = libver
                super().__init__(filename, mode=mode, swmr=swmr, **kwargs)
            if old_file_locking is None:
                del os.environ["HDF5_USE_FILE_LOCKING"]
            else:
                os.environ["HDF5_USE_FILE_LOCKING"] = old_file_locking

    @contextmanager
    def _protect_init(self, filename):
        """Makes sure no other file is opened/created
        or protected sections associated to the filename
        are executed.
        """
        lockname = os.path.abspath(filename)
        with self._LOCKPOOL.acquire(None):
            with self._LOCKPOOL.acquire(lockname):
                yield

    @contextmanager
    def protect(self):
        """Protected section associated to this file."""
        lockname = os.path.abspath(self.filename)
        with self._LOCKPOOL.acquire(lockname):
            yield


def isErrno(e, errno):
    """
    check if the errno happen

    :param OSError e:
    :returns bool:
    """
    # Because e.__cause__ is None for chained exceptions
    return f"errno = {errno}" in "".join(traceback.format_exc())


def check_virtual_sources_exist(fname, data_path):
    """
    Check that a virtual dataset points to actual data.

    :param str fname: HDF5 file path
    :param str data_path: Path within the HDF5 file

    :return bool res: Whether the virtual dataset points to actual data.
    """
    with HDF5File(fname, "r") as f:
        if data_path not in f:
            _logger.error(f"No dataset {data_path} in file {fname}")
            return False
        dptr = f[data_path]
        if not dptr.is_virtual:
            return True
        for vsource in dptr.virtual_sources():
            vsource_fname = os.path.join(
                os.path.dirname(dptr.file.filename), vsource.file_name
            )
            if not os.path.isfile(vsource_fname):
                _logger.error(f"No such file: {vsource_fname}")
                return False
            elif not check_virtual_sources_exist(vsource_fname, vsource.dset_name):
                _logger.error(f"Error with virtual source {vsource_fname}")
                return False
    return True


def from_data_url_to_virtual_source(url: DataUrl) -> tuple:
    """
    convert a DataUrl to a set (as tuple) of h5py.VirtualSource

    :param DataUrl url: url to be converted to a virtual source. It must target a 2D detector
    :return: (h5py.VirtualSource, tuple(shape of the virtual source), numpy.drype: type of the dataset associated with the virtual source)
    :rtype: tuple
    """
    if not isinstance(url, DataUrl):
        raise TypeError(
            f"url is expected to be an instance of DataUrl and not {type(url)}"
        )

    with HDF5File(url.file_path(), mode="r") as o_h5s:
        original_data_shape = o_h5s[url.data_path()].shape
        data_type = o_h5s[url.data_path()].dtype
        if len(original_data_shape) == 2:
            original_data_shape = (
                1,
                original_data_shape[0],
                original_data_shape[1],
            )

        vs_shape = original_data_shape
        if url.data_slice() is not None:
            vs_shape = (
                url.data_slice().stop - url.data_slice().start,
                original_data_shape[-2],
                original_data_shape[-1],
            )

    vs = h5py.VirtualSource(
        url.file_path(), url.data_path(), shape=vs_shape, dtype=data_type
    )

    if url.data_slice() is not None:
        vs.sel = selection.select(original_data_shape, url.data_slice())
    return vs, vs_shape, data_type


def from_virtual_source_to_data_url(vs: h5py.VirtualSource) -> DataUrl:
    """
    convert a h5py.VirtualSource to a DataUrl

    :param h5py.VirtualSource vs: virtual source to be converted to a DataUrl
    :return: url
    :rtype: DataUrl
    """
    if not isinstance(vs, h5py.VirtualSource):
        raise TypeError(
            f"vs is expected to be an instance of h5py.VirtualSorce and not {type(vs)}"
        )
    url = DataUrl(file_path=vs.path, data_path=vs.name, scheme="silx")
    return url


@contextmanager
def cwd_context(new_cwd=None):
    """
    create a context with 'new_cwd'.

    on entry update current working directory to 'new_cwd' and reset previous 'working_directory' at exit
    :param Optional[str] new_cwd: current working directory to use in the context
    """
    try:
        curdir = os.getcwd()
    except Exception as e:
        _logger.error(e)
        curdir = None
    try:
        if new_cwd is not None and os.path.isfile(new_cwd):
            new_cwd = os.path.dirname(new_cwd)
        if new_cwd not in (None, ""):
            os.chdir(new_cwd)
        yield
    finally:
        if curdir is not None:
            os.chdir(curdir)


def to_target_rel_path(file_path: str, target_path: str) -> str:
    """
    cast file_path to a relative path according to target_path.
    This is used to deduce h5py.VirtualSource path

    :param str file_path: file path to be moved to relative
    :param str target_path: target used as 'reference' to get relative path
    :return: relative path of file_path compared to target_path
    :rtype: str
    """
    file_path = os.path.realpath(file_path)
    target_path = os.path.realpath(target_path)
    path = os.path.relpath(file_path, os.path.dirname(target_path))
    if not path.startswith("./"):
        path = "./" + path
    return path



"""
module for handling a `nxsample <https://manual.nexusformat.org/classes/base_classes/NXsample.html>`_
"""

import logging
from functools import partial
from operator import is_not
from typing import Iterable, Optional, Tuple

import numpy
from silx.utils.proxy import docstring
from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from pyunitsystem.metricsystem import MetricSystem

from nxtomo.nxobject.nxobject import ElementWithUnit, NXobject
from nxtomo.nxobject.nxtransformations import NXtransformations
from nxtomo.utils import cast_and_check_array_1D, get_data, get_data_and_unit

_logger = logging.getLogger(__name__)


class NXsample(NXobject):
    def __init__(self, node_name="sample", parent: Optional[NXobject] = None) -> None:
        """
        representation of `nexus NXsample <https://manual.nexusformat.org/classes/base_classes/NXsample.html>`_.
        A monitor of incident beam data.

        :param str node_name: name of the detector in the hierarchy
        :param Optional[NXObject] parent: parent in the nexus hierarchy
        """
        super().__init__(node_name=node_name, parent=parent)
        self._set_freeze(False)
        self._name = None
        self._rotation_angle = None
        self.rocking = None
        self.n_steps_rocking = None
        self.n_steps_rotation = None
        self._x_translation = ElementWithUnit(default_unit=MetricSystem.METER)
        self._y_translation = ElementWithUnit(default_unit=MetricSystem.METER)
        self._z_translation = ElementWithUnit(default_unit=MetricSystem.METER)
        self._transformations = tuple()
        self._set_freeze(True)

    @property
    def name(self) -> Optional[str]:
        """sample name"""
        return self._name

    @name.setter
    def name(self, name: Optional[str]) -> None:
        if not isinstance(name, (type(None), str)):
            raise TypeError(f"name is expected to be None or str not {type(name)}")
        self._name = name

    @property
    def rotation_angle(self) -> Optional[numpy.ndarray]:
        """sample rotation angle. One per frame"""
        return self._rotation_angle

    @rotation_angle.setter
    def rotation_angle(self, rotation_angle: Optional[Iterable]):
        self._rotation_angle = cast_and_check_array_1D(rotation_angle, "rotation_angle")

    @property
    def x_translation(self) -> Optional[numpy.ndarray]:
        """sample translation along x. See `modelization at esrf <https://tomo.gitlab-pages.esrf.fr/ebs-tomo/master/modelization.html>`_ for more information"""
        return self._x_translation

    @x_translation.setter
    def x_translation(self, x_translation: Optional[Iterable]):
        self._x_translation.value = cast_and_check_array_1D(
            x_translation, "x_translation"
        )

    @property
    def y_translation(self) -> Optional[numpy.ndarray]:
        """sample translation along y. See `modelization at esrf <https://tomo.gitlab-pages.esrf.fr/ebs-tomo/master/modelization.html>`_ for more information"""
        return self._y_translation

    @y_translation.setter
    def y_translation(self, y_translation: Optional[Iterable]):
        self._y_translation.value = cast_and_check_array_1D(
            y_translation, "y_translation"
        )

    @property
    def z_translation(self) -> Optional[numpy.ndarray]:
        """sample translation along z. See `modelization at esrf <https://tomo.gitlab-pages.esrf.fr/ebs-tomo/master/modelization.html>`_ for more information"""
        return self._z_translation

    @z_translation.setter
    def z_translation(self, z_translation: Optional[Iterable]):
        self._z_translation.value = cast_and_check_array_1D(
            z_translation, "z_translation"
        )

    @property
    def transformations(self) -> Tuple[NXtransformations]:
        """detector transformations as `NXtransformations <https://manual.nexusformat.org/classes/base_classes/NXtransformations.html>`_"""
        return self._transformations

    @transformations.setter
    def transformations(self, transformations: Tuple[NXtransformations]):
        if not isinstance(transformations, tuple):
            raise TypeError
        for transformation in transformations:
            if not isinstance(transformation, NXtransformations):
                raise TypeError

    @docstring(NXobject)
    def to_nx_dict(
        self,
        nexus_path_version: Optional[float] = None,
        data_path: Optional[str] = None,
    ) -> dict:
        nexus_paths = get_nexus_paths(nexus_path_version)
        nexus_sample_paths = nexus_paths.nx_sample_paths

        nx_dict = {}

        if self.name is not None:
            path_name = f"{self.path}/{nexus_sample_paths.NAME}"
            nx_dict[path_name] = self.name
        if self.rotation_angle is not None:
            path_rotation_angle = f"{self.path}/{nexus_sample_paths.ROTATION_ANGLE}"
            nx_dict[path_rotation_angle] = self.rotation_angle
            nx_dict["@".join([path_rotation_angle, "units"])] = "degree"
        if self.rocking is not None:
            path_rocking = f"{self.path}/{nexus_sample_paths.ROCKING}"
            nx_dict[path_rocking] = self.rocking
        if self.n_steps_rocking is not None:
            path_n_steps_rocking = f"{self.path}/{nexus_sample_paths.N_STEPS_ROCKING}"
            nx_dict[path_n_steps_rocking] = self.n_steps_rocking
        if self.n_steps_rotation is not None:
            path_n_steps_rotation = f"{self.path}/{nexus_sample_paths.N_STEPS_ROTATION}"
            nx_dict[path_n_steps_rotation] = self.n_steps_rotation
        if self.x_translation.value is not None:
            path_x_translation = f"{self.path}/{nexus_sample_paths.X_TRANSLATION}"
            nx_dict[path_x_translation] = self.x_translation.value
            nx_dict["@".join([path_x_translation, "units"])] = str(
                self.x_translation.unit
            )
        if self.y_translation.value is not None:
            path_y_translation = f"{self.path}/{nexus_sample_paths.Y_TRANSLATION}"
            nx_dict[path_y_translation] = self.y_translation.value
            nx_dict["@".join([path_y_translation, "units"])] = str(
                self.y_translation.unit
            )
        if self.z_translation.value is not None:
            path_z_translation = f"{self.path}/{nexus_sample_paths.Z_TRANSLATION}"
            nx_dict[path_z_translation] = self.z_translation.value
            nx_dict["@".join([path_z_translation, "units"])] = str(
                self.z_translation.unit
            )

        if nx_dict != {}:
            nx_dict[f"{self.path}@NX_class"] = "NXsample"
        return nx_dict

    def _load(self, file_path: str, data_path: str, nexus_version: float) -> NXobject:
        """
        Create and load an NXsample from data on disk
        """
        nexus_paths = get_nexus_paths(nexus_version)
        nexus_sample_paths = nexus_paths.nx_sample_paths

        self.name = get_data(
            file_path=file_path,
            data_path="/".join([data_path, nexus_sample_paths.NAME]),
        )
        self.rotation_angle, angle_unit = get_data_and_unit(
            file_path=file_path,
            data_path="/".join([data_path, nexus_sample_paths.ROTATION_ANGLE]),
            default_unit="degree",
        )
        if angle_unit == "degree":
            pass
        elif isinstance(angle_unit, str) and angle_unit.lower() in ("rad", "radian"):
            self.rotation_angle = numpy.rad2deg(self.rotation_angle)
        elif angle_unit is not None:
            raise ValueError(f"rotation angle unit not recognized: {angle_unit}")

        self.x_translation, self.x_translation.unit = get_data_and_unit(
            file_path=file_path,
            data_path="/".join([data_path, nexus_sample_paths.X_TRANSLATION]),
            default_unit=MetricSystem.METER,
        )
        self.y_translation, self.y_translation.unit = get_data_and_unit(
            file_path=file_path,
            data_path="/".join([data_path, nexus_sample_paths.Y_TRANSLATION]),
            default_unit=MetricSystem.METER,
        )
        self.z_translation, self.z_translation.unit = get_data_and_unit(
            file_path=file_path,
            data_path="/".join([data_path, nexus_sample_paths.Z_TRANSLATION]),
            default_unit=MetricSystem.METER,
        )

    @staticmethod
    @docstring(NXobject)
    def concatenate(nx_objects: tuple, node_name="sample"):
        nx_objects = tuple(filter(partial(is_not, None), nx_objects))
        # filter None obj
        if len(nx_objects) == 0:
            return None
        # warning: later we make the assumption that nx_objects contains at least one element
        for nx_obj in nx_objects:
            if not isinstance(nx_obj, NXsample):
                raise TypeError("Cannot concatenate non NXsample object")

        nx_sample = NXsample(node_name)
        _logger.info(f"sample name {nx_objects[0].name} will be picked")
        nx_sample.name = nx_objects[0].name

        rotation_angles = [
            nx_obj.rotation_angle
            for nx_obj in nx_objects
            if nx_obj.rotation_angle is not None
        ]
        if len(rotation_angles) > 0:
            nx_sample.rotation_angle = numpy.concatenate(rotation_angles)

        x_translations = [
            nx_obj.x_translation.value * nx_obj.x_translation.unit.value
            for nx_obj in nx_objects
            if (
                nx_obj.x_translation is not None
                and nx_obj.x_translation.value is not None
            )
        ]
        if len(x_translations) > 0:
            nx_sample.x_translation = numpy.concatenate(x_translations)

        y_translations = [
            nx_obj.y_translation.value * nx_obj.y_translation.unit.value
            for nx_obj in nx_objects
            if (
                nx_obj.y_translation is not None
                and nx_obj.y_translation.value is not None
            )
        ]
        if len(y_translations) > 0:
            nx_sample.y_translation = numpy.concatenate(y_translations)

        z_translations = [
            nx_obj.z_translation.value * nx_obj.z_translation.unit.value
            for nx_obj in nx_objects
            if (
                nx_obj.z_translation is not None
                and nx_obj.z_translation.value is not None
            )
        ]
        if len(z_translations) > 0:
            nx_sample.z_translation = numpy.concatenate(z_translations)

        rocking_list = list(
            filter(
                partial(is_not, None),
                [nx_obj.rocking for nx_obj in nx_objects],
            )
        )
        if len(rocking_list) > 0:
            nx_sample.rocking = numpy.concatenate(rocking_list)

        return nx_sample



"""nexus path used to define a `NXsample <https://manual.nexusformat.org/classes/base_classes/NXsample.html>`_"""

from . import nxtransformations


class NEXUS_SAMPLE_PATH:
    NAME = "sample_name"

    ROTATION_ANGLE = "rotation_angle"

    X_TRANSLATION = "x_translation"

    Y_TRANSLATION = "y_translation"

    Z_TRANSLATION = "z_translation"

    ROCKING = "rocking"

    BASE_TILT = "base_tilt"

    N_STEPS_ROCKING = "n_step_rocking"

    N_STEPS_ROTATION = "n_step_rotation"

    NX_TRANSFORMATIONS = None

    NX_TRANSFORMATIONS_PATHS = None


class NEXUS_SAMPLE_PATH_V_1_0(NEXUS_SAMPLE_PATH):
    pass


class NEXUS_SAMPLE_PATH_V_1_1(NEXUS_SAMPLE_PATH_V_1_0):
    NAME = "name"


class NEXUS_SAMPLE_PATH_V_1_2(NEXUS_SAMPLE_PATH_V_1_1):
    pass


class NEXUS_SAMPLE_PATH_V_1_3(NEXUS_SAMPLE_PATH_V_1_2):
    NX_TRANSFORMATIONS = "transformations"

    NX_TRANSFORMATIONS_PATHS = nxtransformations.NEXUS_TRANSFORMATIONS_PATH_V_1_3



import os
from datetime import datetime
from tempfile import TemporaryDirectory
import numpy

import pytest
from silx.io.url import DataUrl

from nxtomo.application.nxtomo import NXtomo, copy_nxtomo_file
from nxtomo.nxobject.nxdetector import FieldOfView, ImageKey
from nxtomo.nxobject.utils import concatenate


nexus_path_versions = (1.1, 1.0, None)


@pytest.mark.parametrize("nexus_path_version", nexus_path_versions)
def test_nx_tomo(nexus_path_version):
    nx_tomo = NXtomo(node_name="")

    # check start time
    with pytest.raises(TypeError):
        nx_tomo.start_time = 12
    nx_tomo.start_time = datetime.now()

    # check end time
    with pytest.raises(TypeError):
        nx_tomo.end_time = 12
    nx_tomo.end_time = datetime(2022, 2, 27)

    # check sample
    with pytest.raises(TypeError):
        nx_tomo.sample = "tata"

    # check detector
    with pytest.raises(TypeError):
        nx_tomo.instrument.detector = "tata"

    # check energy
    with pytest.raises(TypeError):
        nx_tomo.energy = "tata"
    nx_tomo.energy = 12.3

    # check group size
    with pytest.raises(TypeError):
        nx_tomo.group_size = "tata"
    nx_tomo.group_size = 3

    # check title
    with pytest.raises(TypeError):
        nx_tomo.title = 12
    nx_tomo.title = "title"

    # check instrument
    with pytest.raises(TypeError):
        nx_tomo.instrument = "test"

    # check we can't set undefined attributes
    with pytest.raises(AttributeError):
        nx_tomo.test = 12

    # create detector for test
    projections = numpy.random.random(100 * 100 * 8).reshape([8, 100, 100])
    flats_1 = numpy.random.random(100 * 100 * 2).reshape([2, 100, 100])
    darks = numpy.random.random(100 * 100 * 3).reshape([3, 100, 100])
    flats_2 = numpy.random.random(100 * 100 * 2).reshape([2, 100, 100])
    alignment = numpy.random.random(100 * 100 * 1).reshape([1, 100, 100])
    nx_tomo.instrument.detector.data = numpy.concatenate(
        [
            darks,
            flats_1,
            projections,
            flats_2,
            alignment,
        ]
    )

    nx_tomo.instrument.detector.image_key_control = numpy.concatenate(
        [
            [ImageKey.DARK_FIELD] * 3,
            [ImageKey.FLAT_FIELD] * 2,
            [ImageKey.PROJECTION] * 8,
            [ImageKey.FLAT_FIELD] * 2,
            [ImageKey.ALIGNMENT] * 1,
        ]
    )

    nx_tomo.instrument.detector.x_pixel_size = (
        nx_tomo.instrument.detector.y_pixel_size
    ) = 1e-7
    nx_tomo.instrument.detector.distance = 0.2
    nx_tomo.instrument.detector.field_of_view = FieldOfView.HALF
    nx_tomo.instrument.detector.count_time = numpy.concatenate(
        [
            [0.2] * 3,  # darks
            [0.1] * 2,  # flats 1
            [0.1] * 8,  # projections
            [0.1] * 2,  # flats 2
            [0.1] * 1,  # alignment
        ]
    )

    # create sample for test
    nx_tomo.sample.name = "my sample"
    nx_tomo.sample.rotation_angle = numpy.concatenate(
        [
            [0.0] * 3,  # darks
            [0.0] * 2,  # flats 1
            numpy.linspace(0, 180, num=8, endpoint=False),  # projections
            [180.0] * 2,  # flats 2
            [0.0],  # alignement
        ]
    )

    n_frames = 3 + 2 + 8 + 2 + 1
    nx_tomo.sample.x_translation = [0.6] * n_frames
    nx_tomo.sample.y_translation = [0.2] * n_frames
    nx_tomo.sample.z_translation = [0.1] * n_frames
    assert nx_tomo.is_root is True
    assert nx_tomo.instrument.is_root is False
    assert (
        nx_tomo.root_path
        == nx_tomo.instrument.root_path
        == nx_tomo.instrument.detector.root_path
    )

    NXtomo.check_consistency(nx_tomo=nx_tomo, raises_error=True)

    with TemporaryDirectory() as folder:
        file_path = os.path.join(folder, "nexus_file.hdf5")

        nx_tomo.save(
            file_path=file_path,
            data_path="entry",
            nexus_path_version=nexus_path_version,
        )
        assert os.path.exists(file_path)

        # insure we can read it back
        scan = NXtomo().load(file_path, data_path="entry")
        assert (
            len(
                tuple(
                    filter(
                        lambda image_key: image_key is ImageKey.FLAT_FIELD,
                        scan.instrument.detector.image_key_control,
                    )
                )
            )
            == 4
        )
        assert (
            len(
                tuple(
                    filter(
                        lambda image_key: image_key is ImageKey.DARK_FIELD,
                        scan.instrument.detector.image_key_control,
                    )
                )
            )
            == 3
        )
        assert (
            len(
                tuple(
                    filter(
                        lambda image_key: image_key is ImageKey.PROJECTION,
                        scan.instrument.detector.image_key_control,
                    )
                )
            )
            == 8
        )
        assert (
            len(
                tuple(
                    filter(
                        lambda image_key: image_key is ImageKey.ALIGNMENT,
                        scan.instrument.detector.image_key_control,
                    )
                )
            )
            == 1
        )
        assert scan.energy.value == 12.3
        assert scan.instrument.detector.x_pixel_size.value == 1e-7
        assert scan.instrument.detector.y_pixel_size.value == 1e-7
        assert scan.instrument.detector.distance.value == 0.2
        assert scan.instrument.detector.field_of_view == FieldOfView.HALF
        assert scan.sample.name == "my sample"
        assert (
            len(scan.sample.x_translation.value)
            == len(scan.sample.y_translation.value)
            == len(scan.sample.z_translation.value)
            == n_frames
        )
        assert scan.sample.x_translation.value[0] == 0.6
        assert scan.sample.y_translation.value[0] == 0.2
        assert scan.sample.z_translation.value[0] == 0.1
        if nexus_path_version != 1.0:
            assert scan.instrument.source.name is not None
            assert scan.instrument.source.type is not None

        # try to load it from the disk
        loaded_nx_tomo = NXtomo("test").load(file_path=file_path, data_path="entry")
        assert isinstance(loaded_nx_tomo, NXtomo)
        assert loaded_nx_tomo.energy.value == nx_tomo.energy.value
        assert loaded_nx_tomo.energy.unit == nx_tomo.energy.unit
        assert loaded_nx_tomo.start_time == nx_tomo.start_time
        assert loaded_nx_tomo.end_time == nx_tomo.end_time
        assert (
            loaded_nx_tomo.instrument.detector.x_pixel_size.value
            == nx_tomo.instrument.detector.x_pixel_size.value
        )
        assert (
            loaded_nx_tomo.instrument.detector.x_pixel_size.unit
            == nx_tomo.instrument.detector.x_pixel_size.unit
        )
        assert (
            loaded_nx_tomo.instrument.detector.y_pixel_size.value
            == nx_tomo.instrument.detector.y_pixel_size.value
        )
        assert (
            loaded_nx_tomo.instrument.detector.field_of_view
            == nx_tomo.instrument.detector.field_of_view
        )
        numpy.testing.assert_array_equal(
            loaded_nx_tomo.instrument.detector.count_time.value,
            nx_tomo.instrument.detector.count_time.value,
        )
        assert (
            loaded_nx_tomo.instrument.detector.count_time.unit
            == nx_tomo.instrument.detector.count_time.unit
        )
        assert (
            loaded_nx_tomo.instrument.detector.distance.value
            == nx_tomo.instrument.detector.distance.value
        )
        assert (
            loaded_nx_tomo.instrument.detector.distance.unit
            == nx_tomo.instrument.detector.distance.unit
        )
        assert (
            loaded_nx_tomo.instrument.detector.estimated_cor_from_motor
            == nx_tomo.instrument.detector.estimated_cor_from_motor
        )

        numpy.testing.assert_array_equal(
            loaded_nx_tomo.instrument.detector.image_key_control,
            nx_tomo.instrument.detector.image_key_control,
        )
        numpy.testing.assert_array_equal(
            loaded_nx_tomo.instrument.detector.image_key,
            nx_tomo.instrument.detector.image_key,
        )

        assert loaded_nx_tomo.sample.name == nx_tomo.sample.name
        assert loaded_nx_tomo.sample.rotation_angle is not None
        numpy.testing.assert_array_almost_equal(
            loaded_nx_tomo.sample.rotation_angle, nx_tomo.sample.rotation_angle
        )
        numpy.testing.assert_array_almost_equal(
            loaded_nx_tomo.sample.x_translation.value,
            nx_tomo.sample.x_translation.value,
        )
        numpy.testing.assert_array_almost_equal(
            loaded_nx_tomo.sample.y_translation.value,
            nx_tomo.sample.y_translation.value,
        )
        numpy.testing.assert_array_almost_equal(
            loaded_nx_tomo.sample.z_translation.value,
            nx_tomo.sample.z_translation.value,
        )

        loaded_nx_tomo = NXtomo("test").load(
            file_path=file_path, data_path="entry", detector_data_as="as_numpy_array"
        )
        numpy.testing.assert_array_almost_equal(
            loaded_nx_tomo.instrument.detector.data,
            nx_tomo.instrument.detector.data,
        )
        loaded_nx_tomo = NXtomo("test").load(
            file_path=file_path, data_path="entry", detector_data_as="as_data_url"
        )
        assert isinstance(loaded_nx_tomo.instrument.detector.data[0], DataUrl)
        with pytest.raises(ValueError):
            # check an error is raise because the dataset is not virtual
            loaded_nx_tomo = NXtomo("test").load(
                file_path=file_path,
                data_path="entry",
                detector_data_as="as_virtual_source",
            )

        # test concatenation
        nx_tomo_concat = concatenate([loaded_nx_tomo, None, loaded_nx_tomo])
        concat_file = os.path.join(folder, "concatenated_nexus_file.hdf5")

        nx_tomo_concat.save(
            file_path=concat_file,
            data_path="myentry",
            nexus_path_version=nexus_path_version,
        )
        loaded_concatenated_nx_tomo = NXtomo("test").load(
            file_path=concat_file,
            data_path="myentry",
            detector_data_as="as_virtual_source",
        )
        numpy.testing.assert_array_almost_equal(
            loaded_concatenated_nx_tomo.sample.rotation_angle,
            numpy.concatenate(
                [
                    nx_tomo.sample.rotation_angle,
                    nx_tomo.sample.rotation_angle,
                ]
            ),
        )
        numpy.testing.assert_array_almost_equal(
            loaded_concatenated_nx_tomo.sample.x_translation.value,
            numpy.concatenate(
                [
                    nx_tomo.sample.x_translation.value,
                    nx_tomo.sample.x_translation.value,
                ]
            ),
        )

        with pytest.raises(TypeError):
            concatenate([1, 2])


@pytest.mark.parametrize("nexus_path_version", nexus_path_versions)
def test_nx_tomo_subselection(nexus_path_version):
    """
    test sub_select_from_projection_angle_range
    """
    nx_tomo = NXtomo()
    nx_tomo.energy = 12.3
    shape = (12, 12)
    data_dark = numpy.ones(shape)
    data_flat = numpy.ones(shape) * 2.0
    data_projection = numpy.ones(shape) * 3.0
    str(nx_tomo)
    nx_tomo.instrument.detector.data = numpy.concatenate(
        (
            data_dark,
            data_dark,
            data_flat,
            data_projection,
            data_projection,
            data_projection,
            data_flat,
            data_projection,
            data_projection,
            data_projection,
            data_flat,
        )
    )
    nx_tomo.instrument.detector.image_key_control = numpy.array(
        (
            ImageKey.DARK_FIELD,
            ImageKey.DARK_FIELD,
            ImageKey.FLAT_FIELD,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.FLAT_FIELD,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.FLAT_FIELD,
        )
    )

    original_angles = numpy.array(
        (
            0,
            0,
            0,
            10,
            20.5,
            22.5,
            180,
            180,
            200,
            300.2,
            300.2,
        )
    )
    nx_tomo.sample.rotation_angle = original_angles

    nx_tomo_sub_1 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=10,
        angle_interval=12.5,
        shift_angles=False,
    )

    numpy.testing.assert_equal(
        nx_tomo_sub_1.instrument.detector.image_key_control,
        numpy.array(
            (
                ImageKey.DARK_FIELD,
                ImageKey.DARK_FIELD,
                ImageKey.FLAT_FIELD,
                ImageKey.INVALID,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
                ImageKey.INVALID,
                ImageKey.INVALID,
                ImageKey.INVALID,
                ImageKey.FLAT_FIELD,
            )
        ),
    )
    numpy.testing.assert_equal(
        nx_tomo_sub_1.sample.rotation_angle,
        original_angles,
    )

    nx_tomo_sub_2 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=10,
        angle_interval=20,
        shift_angles=True,
    )

    numpy.testing.assert_equal(
        nx_tomo_sub_2.sample.rotation_angle[0:3],
        0.0,
    )
    numpy.testing.assert_array_equal(
        nx_tomo_sub_2.sample.rotation_angle[3:6],
        numpy.array([0.0, 10.5, 12.5]),
    )

    nx_tomo_sub_3 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=-10,
        angle_interval=300,
        shift_angles=False,
    )

    numpy.testing.assert_equal(
        nx_tomo_sub_3.instrument.detector.image_key_control,
        numpy.array(
            (
                ImageKey.DARK_FIELD,
                ImageKey.DARK_FIELD,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.INVALID,
                ImageKey.FLAT_FIELD,
            )
        ),
    )

    nx_tomo_sub_4 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=-10,
        angle_interval=None,
        shift_angles=False,
    )
    numpy.testing.assert_equal(
        nx_tomo_sub_4.instrument.detector.image_key_control,
        numpy.array(
            (
                ImageKey.DARK_FIELD,
                ImageKey.DARK_FIELD,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
            )
        ),
    )


def test_bliss_original_files(tmp_path):
    """
    test about NXtomo.bliss_original_files
    """
    test_dir = tmp_path / "test_bliss_original_files"
    test_dir.mkdir()

    nx_tomo_1 = NXtomo()
    with pytest.raises(TypeError):
        nx_tomo_1.bliss_original_files = 12

    nx_tomo_1.bliss_original_files = ("/path/1", "/path/2")

    nx_tomo_2 = NXtomo()
    nx_tomo_2.bliss_original_files = ("/path/2", "/path/3")

    nx_tomo_3 = NXtomo()

    nx_tomo_4 = NXtomo()
    nx_tomo_4.bliss_original_files = ()

    nx_tomo_concat = concatenate([nx_tomo_1, nx_tomo_2, nx_tomo_3])
    assert nx_tomo_concat.bliss_original_files == ("/path/1", "/path/2", "/path/3")

    output_nx_tomo_concat = os.path.join(test_dir, "nx_concat.nx")
    nx_tomo_concat.save(output_nx_tomo_concat, "/entry_concat")

    loaded_nx_tomo = NXtomo().load(output_nx_tomo_concat, "/entry_concat")
    assert loaded_nx_tomo.bliss_original_files == ("/path/1", "/path/2", "/path/3")

    output_nx_tomo_file = os.path.join(test_dir, "nx_tomo.nx")
    nx_tomo_3.save(output_nx_tomo_file, "/entry0000")
    loaded_nx_tomo = NXtomo().load(output_nx_tomo_file, "/entry0000")
    assert loaded_nx_tomo.bliss_original_files is None

    nx_tomo_4.save(output_nx_tomo_file, "/entry0000", overwrite=True)
    loaded_nx_tomo = NXtomo().load(output_nx_tomo_file, "/entry0000")
    assert loaded_nx_tomo.bliss_original_files == ()


@pytest.mark.parametrize("vds_resolution", ("update", "remove"))
def test_copy_nxtomo_file(tmp_path, vds_resolution):
    """test 'copy_nxtomo_file' function"""
    input_folder = tmp_path / "input"
    input_folder.mkdir()
    input_nx_tomo_file = os.path.join(input_folder, "nexus.nx")

    output_folder = tmp_path / "output"
    output_folder.mkdir()

    nx_tomo = NXtomo()
    nx_tomo.save(input_nx_tomo_file, "/entry0000")

    output_file = os.path.join(output_folder, "nxtomo.nx")

    copy_nxtomo_file(
        input_nx_tomo_file,
        entries=None,
        output_file=output_file,
        vds_resolution=vds_resolution,
    )
    assert os.path.exists(output_file)
